/******************************************************************************
*                                                                             *
* project name:    STARS                                                      *
* initial date:    06/06/2000                                                 *
* authors:         thomas.nussbaumer@gmx.net                                  *
*                  lionel_debroux@yahoo.fr                                    *
* description:     starfield effect demonstration (optimized version)         *
*                                                                             *
*******************************************************************************

* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
* NOTE: depends on TIGCC 0.95 and later to compile
* -mpcrel is buggy in the first releases of TIGCC 0.95
* TIGCC 0.95 Beta 4 and newer versions include the SET_FILE_IN_USE_BIT patch
* more than they should...
* !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

-------------------------------------------------------------------------------
 Revision History                                                            
-------------------------------------------------------------------------------
v2.10: 07/06/2004
    Important bugfix: the program could crash with "Protected Memory Violation"
    under some conditions.
    Indeed, if the program was started with a number of stars different from
    MAX_NUMBER_STARS, there were graphical glitches, and when increasing the
    number of stars to MAX_NUMBER_STARS, the calculator crashed with
    "Protected Memory Violation". This was due to a bug (present since the
    number of stars can be modified) in the amount of memory allocated
    (number_stars instead of MAX_NUMBER_STARS, 'number_stars' was only
    correct when the number of stars was a constant): the table of stars was
    not fully cleared...
    
    Optimizations by using bit instructions, optimizing structure star_type,
    rearranging code, using gotos, auxiliary pointers instead of repeated
    array dereferences...
    
    Made help screen optional (disabled by default).

v2.00: 12/02/2002
    by Lionel DEBROUX, lionel_debroux@yahoo.fr

    Many modifications...
    Now it's a "true" screensaver (actually battery-waster, screensavers are
    *very* useless on LCDs).
    Set default contrast value to make the screensaver look better,
    whatever the contrast is. Anyway, you can change it with LEFT or RIGHT.
    Also, the speed of the starfield can be changed with ( and ). Finally, I
    managed to implement the 'backwards travel'...

    Credits go to Greg Dietsche.

    size optimizations using -fomit-frame-pointer and static keyword for
    functions [tnussb]

v1.00: 16/06/2000
       initial version By Thomas NUSSBAUMER

*******************************************************************************/

// Uncomment this definition if you want the help screen at the beginning.
//#define INCLUDE_HELP_SCREEN

// Uncomment one of those to produce the 89, 92+ or V200 version.
//#define USE_TI89
#define USE_TI92P
//#define USE_V200


#if    defined(USE_TI89) && defined(USE_TI92P) \
    || defined(USE_TI89) && defined(USE_V200)
  #error Cannot produce both versions at once
#elif defined(USE_TI89)
  #define C89_92(x,y) (x)
#elif (defined(USE_TI92P) || defined(USE_V200))
  #define USE_TI92P
  #define USE_V200
  #define C89_92(x,y) (y)
#else
  #error Please specify either #define USE_TI89 or #define USE_TI92P or \
  #define USE_V200 right above
#endif

#define SAVE_SCREEN           // Save/Restore LCD Contents

#define OPTIMIZE_ROM_CALLS  	// Use ROM Call Optimization
#define NO_EXIT_SUPPORT
#define MIN_AMS 100
#define NO_AMS_CHECK

#include <tigcclib.h>

#define MY_LCD_WIDTH  C89_92(160,240)
#define MY_LCD_HEIGHT C89_92(100,128)


/*---------------------------------------------------------------------------*/
/* some constants defining the look of the starfield                         */
/*---------------------------------------------------------------------------*/
#define XOFFSET      (MY_LCD_WIDTH>>1)
#define YOFFSET      (MY_LCD_HEIGHT>>1)
#define ZOFFSET      254
#define MAX_STAR_X   (RAND_MAX)
#define MAX_STAR_Y   (RAND_MAX)
#define MAX_STARX2   (MAX_STAR_X>>1)
#define MAX_STARY2   (MAX_STAR_Y>>1)
#define BORDER_BLACK 90
#define BORDER_GRAY  140


#define MAX_NUMBER_STARS        C89_92(750,1000)   // number of maximum stars on screen
#define MIN_NUMBER_STARS        C89_92(100,140)    // number of minimum stars on screen
#define NUMBER_STARS_INCREMENT  20                 // number of stars changed by each keypress

#define MAX_ZDECREMENT            9   // maximum "speed" in forward direction
#define MIN_ZDECREMENT           -9   // maximum "speed" in backward direction


/*---------------------------------------------------------------------------*/
/* structure which stores a star's location                                  */
/*---------------------------------------------------------------------------*/
typedef struct {
    short offset;  // offset of old location for cleanup
    unsigned char bitnumber;    // inverted mask of old location for cleanup
    unsigned char z;
    short x,y;
} star_type;


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/
star_type* stars = NULL;  // array of stars

static short number_stars = C89_92(500,750);
static short zdecrement   = 4;

/*===========================================================================*/
/* initializes star with given index                                         */
/*===========================================================================*/
static inline void CreateStar(star_type *ptr) {
    ptr->x = random(RAND_MAX)-(RAND_MAX>>1);
    ptr->y = random(RAND_MAX)-(RAND_MAX>>1);
    if (zdecrement > 0) {
        ptr->z = ZOFFSET;
    }
    else {
        ptr->z = random(ZOFFSET-10)+1;
    }
}


/*===========================================================================*/
/* delete star at old position/calculate new position/draw new position      */
/*===========================================================================*/
static inline void DrawStars(void) {
    short         offset;
    unsigned char bitnumber;
    short         drawx;
    short         drawy;
    short         i;
    star_type    *ptr;
    void         *light;
    void         *dark;

    ptr = &stars[0];
    light = GrayGetPlane(LIGHT_PLANE);
    dark = GrayGetPlane(DARK_PLANE);

    for (i=0;i<number_stars;i++) {
        // offset can't be > 32768, so checking for < 0 is enough.
        if (ptr->offset >= 0) {
            poke_bclr(((unsigned char *)light)+ptr->offset,ptr->bitnumber);
            poke_bclr(((unsigned char *)dark)+ptr->offset,ptr->bitnumber);
            ptr->offset = -1;
        }

        // calculate new drawing x position
        // and check if it out of screen
        drawx  = ptr->x / ptr->z + XOFFSET;
        if (drawx < 0 || drawx >= MY_LCD_WIDTH) {
            goto _CREATESTAR_;
        }

        // calculate new drawing y position
        // and check if it out of screen
        drawy  = ptr->y / ptr->z + YOFFSET;
        if (drawy < 0 || drawy >= MY_LCD_HEIGHT) {
            goto _CREATESTAR_;
        }

        // calculate byte offset of pixel
        // We use this construct, because -Os will generate a mulu.w if we use *30.
        offset = (((((drawy)+(drawy))<<4)-((drawy)+(drawy)))+((drawx)>>3));

        // calculate bit number
        bitnumber = (~(drawx));

        // set pixel in plane 1 and/or plane2
        // depending on z value
        if (ptr->z < BORDER_GRAY) {
            poke_bset(((unsigned char *)dark)+offset,bitnumber);
            if (ptr->z < BORDER_BLACK) {
                poke_bset(((unsigned char *)light)+offset,bitnumber);
            }
        }
        else {
            poke_bset(((unsigned char *)light)+offset,bitnumber);
        }

        // store offset and bit number so that we can remove the pixel easily
        ptr->bitnumber = bitnumber;
        ptr->offset = offset;

        // decrement z and check if we must
        // generate a new star
        ptr->z -= zdecrement;
        if (ptr->z <= 0 || ptr->z > ZOFFSET) {
            _CREATESTAR_:
            CreateStar(ptr);
        }

        ptr++;
    }
}


/*===========================================================================*/
/* Clears 2 gray planes                                                      */
/* Note: using the internal variables directly is a bad practice.            */
/*===========================================================================*/
//void ClearGrayScreen2B(void* lightplane,void* darkplane); // C prototype
void ClearGrayScreen2B(void); // C prototype
asm("xdef ClearGrayScreen2B\n"
"ClearGrayScreen2B:\n"
"      movem.l  %d0-%d7/%a0-%a6,-(%a7)\n"
"      movea.l  __L_plane(%pc),%a0|56(%a7),%a0\n"
"      movea.l  __D_plane(%pc),%a1|60(%a7),%a1\n"
"      lea      (0xf00,%a0),%a0\n"
"      lea      (0xf00,%a1),%a1\n"
"      moveq    #(2*10-1),%d7\n"
"      moveq    #0x0,%d0\n"
"      moveq    #0x0,%d1\n"
"      moveq    #0x0,%d2\n"
"      moveq    #0x0,%d3\n"
"      moveq    #0x0,%d4\n"
"      moveq    #0x0,%d5\n"
"      moveq    #0x0,%d6\n"
"      movea.l  %d0,%a2\n"
"      movea.l  %d0,%a3\n"
"      movea.l  %d0,%a4\n"
"      movea.l  %d0,%a5\n"
"      movea.l  %d0,%a6\n"
"0:\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a1)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a0)\n"
"      movem.l  %d0-%d6/%a2-%a6,-(%a1)\n"
"      dbf      %d7, 0b\n"
"      movem.l  (%a7)+,%d0-%d7/%a0-%a6\n"
"      rts");


/*===========================================================================*/
/* Alias for the above routine                                               */
/*===========================================================================*/
#define ClearGrayScreen ClearGrayScreen2B


// Declaring the two following functions inline, reduces size.
/*===========================================================================*/
/* number of stars changed -> clear the screen and restart timer             */
/*===========================================================================*/
static inline void NrStarsChanged(void) {
    ClearGrayScreen();
}

/*===========================================================================*/
/* wait a while for key debouncing                                           */
/*===========================================================================*/
static inline void debounce(void) {
    asm("moveq #-1,%%d0; 0: dbf %%d0,0b" : : : "d0");
    OSTimerRestart(APD_TIMER);
}


/*===========================================================================*/
/* Returns a pointer to the adress where is stored the contrast.             */
/* This function is from TI-Chess.                                           */
/*===========================================================================*/
/*volatile*/ unsigned char *contrast(void) {
    /*volatile*/ unsigned char *cnt;
    asm("contrast: moveq #4,%%d0;trap #9;" : "=g"(cnt): : "d0");
    return cnt;
}


// Old-fashioned _rowread, leftover from TIGCC 0.93 and before...
#define ARROWS_ROW       ~(1<<0)
#define ESC_ROW          ~C89_92(1<<6,1<<8)
#define OPEN_PARENTH_ROW ~(1<<4)
#define CLOS_PARENTH_ROW ~C89_92(1<<3,1<<4)

#define UP           C89_92(1<<0,1<<5)
#define DOWN         C89_92(1<<2,1<<7)
#define LEFT         C89_92(1<<1,1<<4)
#define RIGHT        C89_92(1<<3,1<<6)
#define ESC          C89_92(1<<0,1<<6)
#define OPEN_PARENTH C89_92(1<<4,1<<5)
#define CLOS_PARENTH C89_92(1<<4,1<<6)


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
    short          i, f = FontGetSys();
    /*volatile*/ unsigned char contsave = 0;
    unsigned char  row = 0;
    INT_HANDLER    save_int1;

#ifdef INCLUDE_HELP_SCREEN
    // show help screen
    ClrScr();
    FontSetSys(F_8x10);
    DrawStr(C89_92(15,55),15,"Starfield Effect",A_NORMAL);
    DrawStr(C89_92(33,73),30,"\xA9 TICT 2004",A_NORMAL);
    FontSetSys(F_4x6);
    DrawStr(C89_92(15,55),50,"UP / DOWN        modify number of stars",A_NORMAL);
    DrawStr(C89_92(15,55),58,"LEFT / RIGHT    modify constrast",A_NORMAL);
    DrawStr(C89_92(15,55),66,"( / )                   modify speed/direction",A_NORMAL);
    ngetchx();
#endif

    // try to allocate memory for stars
    if(!(stars = (star_type*)HeapAllocPtr(sizeof(star_type)*MAX_NUMBER_STARS))) {
        ClrScr();
        ST_helpMsg("ERROR: not enough memory");
        ngetchx();
        return;
    }

    // redirect default interrupt-1 handler to dummy handler
    save_int1=GetIntVec(AUTO_INT_1);
    SetIntVec(AUTO_INT_1,DUMMY_HANDLER);

    // turn on grayscales
    if (!GrayOn()) {
        SetIntVec(AUTO_INT_1,save_int1);
        return;
    }

    // setup initial star positions
    {
        star_type *ptr=&stars[0];
        for (i=0;i<MAX_NUMBER_STARS;i++) {
            ptr->x = random(MAX_STAR_X)-(MAX_STARX2);
            ptr->y = random(MAX_STAR_X)-(MAX_STARY2);
            ptr->z = random(ZOFFSET-1)+1;
            ptr->offset = -1;
            ptr++;
        }
    }

    // clear both planes
    ClearGrayScreen();

    // get initial contrast
    contsave=*contrast();

    //---------------------------------------------
    // main loop
    //---------------------------------------------
    while (1) {
        DrawStars();
        row = _rowread(ARROWS_ROW);
        if (row&UP && (number_stars <= (MAX_NUMBER_STARS - NUMBER_STARS_INCREMENT))) {  // UP
            number_stars += NUMBER_STARS_INCREMENT;
            NrStarsChanged();
            debounce();
        }
        else if (row&DOWN && (number_stars >= (MIN_NUMBER_STARS + NUMBER_STARS_INCREMENT))) {  // DOWN
            number_stars -= NUMBER_STARS_INCREMENT;
            NrStarsChanged();
            debounce();
        }
        else if (row&RIGHT) {  // RIGHT
            OSContrastUp();
            debounce();
        }
        else if (row&LEFT) {  // LEFT
            OSContrastDn();
            debounce();
        }
        else if (_rowread(OPEN_PARENTH_ROW)&OPEN_PARENTH && (zdecrement>MIN_ZDECREMENT)) {  // (
            zdecrement--;
            if (!zdecrement) zdecrement--;
            debounce();
        }
        else if (_rowread(CLOS_PARENTH_ROW)&CLOS_PARENTH && (zdecrement<MAX_ZDECREMENT)) {  // )
            zdecrement++;
            if (!zdecrement) zdecrement++;
            debounce();
        }
        else if (_rowread(ESC_ROW)&ESC) {  // ESC
            GrayOff();

            // Writing to the contrast address does not change the contrast,
            // we have to call OSContrastDn or OSContrastUp...
            *contrast() = contsave+1;
            OSContrastDn();
            GKeyFlush();

            // cleanup and bye =8-)
            HeapFreePtr(stars);
            SetIntVec(AUTO_INT_1,save_int1);
            FontSetSys(f);
            return;
        }
        if (OSTimerExpired(APD_TIMER)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
    }
}
